%%% Generates A matrices and simulates pricing experiments to estimate A.
%%% Calculates how many experiments are needed to obtain a certain
%%% epsilon-delta accuracy criterion, for various numbers of experiments.
%%% Produces a plot of (# of products) vs (# of experiments needed to
%%% achieve uniform epsilon-accuracy).


% how to sparsify: 2 - sparsity, 3 - bounded influence
flag = 2;

% array of # of products to try
narray = 10:10:70;
num_n = size(narray,2);
n_max = narray(num_n);

% sparsity bound used for certain ways of generating parameters
sparsity = min(narray,max(1,round(1.1178124 + 0.6614410*narray ...
            - 0.0068511*narray.^2 + normrnd(0,1,1,num_n))));
for i=2:num_n
    if sparsity(i) < sparsity(i-1)
        sparsity(i) = sparsity(i-1);
    end
end
b = 5;
d = 5;
bound = [b sparsity];

% the desired accuracy criterion
max_epsilon = 1.5;
delta = 0.1;

% bound on noise
c = 0.5;
gaussian_noise = false;
if gaussian_noise
    noise_type = 'noise: N(0,c^2)';
else
    noise_type = 'noise: U[-c,c]';
end

% the number of parallel simulations to run for each instance
n_iters = 10;
% the number of replications of the entire simulation
a_iters = 5;

% randomly generate X's from U[-lambda,lambda]
lambda = 0.5;

vars = 2;

% store the number of experiments needed to achieve uniform
% epsilon-accuracy here
s_needed_sparse = zeros(a_iters,num_n);
s_needed_dense = zeros(a_iters,num_n);

for a=1:a_iters
    % generate the full matrix, bounded in [-b,b]
    [~, ~, A_full, ~, ~] = ...
        generate_parameters(n_max, BOUNDED_ELEMENTS, bound);

    nindex = 0;
    for n=narray
        % display(n);
        nindex = nindex + 1;

        % take only the first n elements of the parameters
        An_full = A_full(1:n,1:n);
        An = sparsify(An_full, sparsity(nindex));

        % the maximum # of experiments to try before giving up
        s_max = 100*n;

        Yn = zeros(n,n,n_iters);
        Yn_full = zeros(n,n,n_iters);

        % set to true if we find an s that satisfies the (epsilon,delta)
        % condition. means that this value of n is done.
        sparse_done = false;
        dense_done = false;

        for s=1:s_max
            done = zeros(vars,n_iters);
            for i=1:n_iters
                % generate pricing decisions
                X = rand(n, 1)*2*lambda - lambda;
                % generate noise
                if gaussian_noise
                    noise = randn(n, 1)*c;
                else
                    noise = rand(n, 1)*2*c - c;
                end

                if ~sparse_done
                    % calculate demand
                    Q = An*X + noise;
                    % update statistic
                    Yn(:,:,i) = Yn(:,:,i) + 3/lambda^2*Q*X';
                    % calculate estimate of A
                    An_hat = Yn(:,:,i)./s;
                    % derive max error
                    maxerror = max(max(abs(An_hat - An)));
                    % if it's less than the desired criterion, we're done
                    if maxerror <= max_epsilon
                        done(1,i) = 1;
                    end
                end
                if ~dense_done
                    % do the same thing for the dense matrix
                    Q_full = An_full*X + noise;
                    Yn_full(:,:,i) = Yn_full(:,:,i) + 3/lambda^2*Q_full*X';
                    An_full_hat = Yn_full(:,:,i)./s;
                    maxerror = max(max(abs(An_full_hat - An_full)));
                    if maxerror <= max_epsilon
                        done(2,i) = 1;
                    end
                end
            end

            % if no more than delta of the parallel instances have
            % satisfied uniform epsilon accuracy, we're done
            if sum(done(1,:))/n_iters >= 1-delta && sparse_done == false
                sparse_done = true;
                s_needed_sparse(a,nindex) = s;
            end
            if sum(done(2,:))/n_iters >= 1-delta && dense_done == false
                dense_done = true;
                s_needed_dense(a,nindex) = s;
            end
                        
            if sparse_done && dense_done
                break;
            end
        end
    end
end

% calculate CIs
tval = tinv(.975,a_iters-1);
s_sparse_mean = mean(s_needed_sparse);
s_sparse_std_err = std(s_needed_sparse);
sparse_CI_width = tval*s_sparse_std_err/sqrt(a_iters);
s_dense_mean = mean(s_needed_dense);
s_dense_std_err = std(s_needed_dense);
dense_CI_width = tval*s_dense_std_err/sqrt(a_iters);

figure
clf
hold on
grid off
errorbar(narray, s_sparse_mean, sparse_CI_width, 'bx', 'MarkerSize', 3);
errorbar(narray, s_dense_mean, dense_CI_width, 'ro', 'MarkerSize', 3);
legend('sparse A', 'dense A', 'Location', 'NorthWest')
xlabel('n: number of products')
ylabel('s: number of experiments')
