%%% Generates A matrices from distributions seeded by parameters obtained
%%% from a past field experiment.
%%%
%%% INPUT
%%% n: number of products
%%% flag: also generates a sparse version of the A matrix, depending on the
%%%       value of the flag
%%%       0 - does not sparsify
%%%       1 - bounds all elements of A to be no more than b in magnitude
%%%       2 - bounds all elements of A to be no more than b in magnitude
%%%           and sets all but k elements of each row to 0
%%%       3 - bounds the L1 norm of each row of A to be no more than d
%%% bound: sparsity parameters
%%%       0 - unused
%%%       1 - [b]
%%%       2 - [b k]
%%%       3 - [d]
%%%
%%% OUTPUT
%%% p: profit-related [unused]
%%% qb: quantity-related [unused]
%%% A: sparse version of matrix
%%% Afull: dense version of matrix
%%% sizes: category size-related [unused]
function [p, qb, A, Afull, sizes] = generate_parameters(n, flag, bound)

%%% CONSTANTS
NONE = 0;
BOUNDED_ELEMENTS = 1;
SPARSITY = 2;
BOUNDED_SPILLOVERS = 3;
SUBCATS = 4;
SUBCATS_INDEP = 5;
MATRIX = 6;

% generate weekly profits using gamma distribution
pmu = 4.4873; %3.7145;
pstdev = 7.3905; %3.7437;
ptheta = pstdev^2 / pmu;
pk = pmu / ptheta;
p = gamrnd(pk, ptheta, n, 1);

% generate baseline demand using gamma distribution
qbmu = 10.72; %2.7029;
qbstdev = 19.57; %8.4425;
qbtheta = qbstdev^2 / qbmu;
qbk = qbmu / qbtheta;
qb = gamrnd(qbk, qbtheta, n, 1);

% statistics for generating subcategory sizes
subcatmu = 24.7275;
subcatstdev = 42.2177;
subcattheta = subcatstdev^2 / subcatmu;
subcatk = subcatmu / subcattheta;

% generate A matrix
% generate diagonals using normal distribution
ownmu = -1.097063981;
ownstdev = 3.23764167;
diags = ownmu + ownstdev.*randn(n,1);
% generate off-diagonals using normal distribution
crossmu = -0.214823293;
crossstdev = -0.214823293;
offdiags = crossmu + crossstdev.*randn(n,n);

% first, create a full matrix (also used for comparison purposes)
Afull = offdiags;
for i=1:n
    % replace diagonal entries
    Afull(i,i) = diags(i);
end

sizes = [];

if flag == NONE
    A = Afull;
elseif flag == BOUNDED_ELEMENTS
    b = bound(1);
    % fix A so that all entries are in the range [-b,b]
    A = min(b.*ones(n), Afull);
    A = max(-b.*ones(n), A);
elseif flag == SPARSITY
    b = bound(1);
    % first bound all elements to the range [-b,b]
    Afull = min(b.*ones(n), Afull);
    Afull = max(-b.*ones(n), Afull);
    % then sparsify the full matrix according to k
    A = sparsify(Afull, ceil(bound(2)));
elseif flag == BOUNDED_SPILLOVERS
    A = zeros(n,n);
    sizes = ones(1,n);
    d = bound(1);
    
    for i=1:n
        % bound the diagonals by d
        A(i,i) = max(-d, min(d, Afull(i,i)));
        
        abssum = 0;
        rand_perm = randperm(n);
        i_index = find(rand_perm == i);
        rand_perm = [rand_perm(1:i_index-1) rand_perm(i_index+1:n)];
        
        for j=1:n-1
            rand_index = rand_perm(j);
            if abssum + abs(Afull(i,rand_index)) < d
                A(i,rand_index) = Afull(i,rand_index);
                abssum = abssum + abs(Afull(i,rand_index));
            else
                half_remaining = (d - abssum)/2;
                A(i,rand_index) = half_remaining*sign(Afull(i,rand_index));
                abssum = abssum + half_remaining;
            end
        end
    end
elseif flag == SUBCATS
    A = zeros(n,n);
    skus = 0;
    sizes = [];
    while skus < n
        new_size = ceil(gamrnd(subcatk, subcattheta));
        new_size = min(new_size, n-skus);
        sizes = [sizes new_size];
        A(skus+1:skus+new_size,skus+1:skus+new_size) = ...
            Afull(skus+1:skus+new_size,skus+1:skus+new_size);
        skus = skus + new_size;
    end
elseif flag == SUBCATS_INDEP
    A = zeros(n,n);
    sizes = zeros(1,n);
    for i=1:n
        subcat_size = min(n, ceil(gamrnd(subcatk, subcattheta)));
        sizes(i) = subcat_size;
        randindices = generate_random_indices(n, subcat_size, i);
        A(i,:) = Afull(i,:).*randindices;
    end
elseif flag == MATRIX
    b = bound(1);
    % first bound all elements to the range [-b,b]
    Afull = min(b.*ones(n), Afull);
    Afull = max(-b.*ones(n), Afull);
    % then sparsify the full matrix according to k total nonzero entries
    A = matrix_sparsify(Afull, ceil(bound(2)));
end
end

function A = matrix_sparsify(Afull, k)
n = size(Afull, 1);    
A = zeros(n,n);

k = min(k, n^2);

indices = randsample(n^2, k);
for i=1:k
    index = indices(i);
    row = floor((index + n - 1)/n);
    col = mod(index-1,n) + 1;
    A(row, col) = Afull(row, col);
end
end